//==================================================================================================
//
// SZpack functions
//
//==================================================================================================
//
// purpose: computation of the SZ signal according to Chluba, Nagai, Sazonov, Nelson, 2012
// 
// comments: - the cluster is assumed to be iso-thermal and internal motions are neglected
//           - computations are performed in the single-scattering approximation
//           - polarization effects are neglected   
//
//==================================================================================================
//
// Author: Jens Chluba (CITA, University of Toronto)
//
// first implementation: May 2012
// last modification: June 2012
//
//==================================================================================================


//==================================================================================================
//
//  Unless stated otherwise the main parameters are:
//  
//  xo          : observer frame photon frequency xo == h nu / k T0 (T0 == todays CMB temperature)
//  Dtau        : line-of-sight scattering optical depth, with explicit use of Lorentz-invariance
//  Te          : temperature of the (cluster frame) electron gas in keV
//  betac       : velocity beta=v/c of the cluster in the CMB frame
//  muc         : direction cosine of the cluster velocity with respect to the line-of-sight in  
//                the CMB rest frame
//  betao       : velocity beta=v/c of the observer in the CMB frame 
//                (from CMB dipole betao=1.241e-3)
//  muo         : direction cosine for the line-of-sight with respect to the observers velocity. 
//                The angle is measured in the observer frame
//
//==================================================================================================

#include <iostream>
#include <string>
#include <fstream>
#include <cmath>
#include <vector>

#include "SZ_Integral.5D.h"
#include "SZ_Integral.3D.h"
#include "SZ_asymptotic.h"
#include "SZ_CNSN_basis.h"
#include "physical_consts.h"

using namespace std;


//==================================================================================================
//
// As discussed by Chluba et el. Nozawa 1998 and 2006 did not explicitly impose the 
// lorentz-invariance of the scattering optical depth. In their convention 
// 
// Dtau --> gammac (1-betac muc) Dtau 
//
// for an observer at rest in the CMB frame. To use this convention call 
// 'use_Nozawa2006_convention();' before excecuting this routines. To reset to the convention 
// of Chluba et al. call 'use_CNSN_convention();' Default is this setting.
//
//==================================================================================================
bool CNSN2012_convention=1;

void use_Nozawa2006_convention(){ CNSN2012_convention=0; }
void use_CNSN_convention(){ CNSN2012_convention=1; }


//==================================================================================================
//
//  Full numerical integration of the 5-dimensional Compton collision term. All terms in betac are
//  included. In principle here one can easily add the scattering of primordial CMB anisotropies.
//
//  eps_Int : relative accuracy for numerical integration (lower than 10^-6 is hard to achieve)
//
//==================================================================================================
double compute_SZ_signal_5D(double xo, 
                            double Dtau, double Te, 
                            double betac, double muc, 
                            double betao, double muo, 
                            double eps_Int)
{
    double The=Te/const_me; // conversion keV --> The = k Te / me c^2
    double gammao=1.0/sqrt(1.0-betao*betao);
    double gammac=1.0/sqrt(1.0-betac*betac);
    
    // transformation of muc to mucc
    double mucc=(muc-betac)/(1.0-betac*muc);
        
    // transformation of xo to xc
    double xc=gammac*(1.0-betac*mucc)*gammao*xo*(1.0+betao*muo);
    
    // change to Nozawa 2006 convention for scattering optical depth
    if(!CNSN2012_convention) Dtau*=gammac*(1.0-betac*mucc);
    
    return Dtau*compute_SZ_distortion_Patterson_5D(xc, The, betac, mucc, eps_Int);
}


//==================================================================================================
//
//  Full numerical integration after reduction of Compton collision term to 3 dimensions. Only terms
//  up to betac^2 are retained. Execution of this routine is significantly faster than for the full 
//  5D integral.
//
//  eps_Int : relative accuracy for numerical integration (lower than 10^-6 is hard to achieve)
//
//==================================================================================================
double compute_SZ_signal_3D(double xo, 
                            double Dtau, double Te, 
                            double betac, double muc, 
                            double betao, double muo, 
                            double eps_Int)
{
    double The=Te/const_me; // conversion keV --> The = k Te / me c^2
    double gammao=1.0/sqrt(1.0-betao*betao);
    double gammac=1.0/sqrt(1.0-betac*betac);
    
    // transformation of muc to mucc
    double mucc=(muc-betac)/(1.0-betac*muc);
    
    // transformation of xo to xc
    double xc=gammac*(1.0-betac*mucc)*gammao*xo*(1.0+betao*muo);
    
    // change to Nozawa 2006 convention for scattering optical depth
    if(!CNSN2012_convention) Dtau*=gammac*(1.0-betac*mucc);

    return Dtau*compute_SZ_distortion_Patterson_3D(xc, The, betac, mucc, "all", eps_Int);
}


//==================================================================================================
//
//  Asymptotic expansion of the Compton collision integral similar to Itoh et al. but up to 
//  10th order in temperature. Kinematic terms are computed according to Chluba et al. 2012.
//
//  The additional parameters are:
//  
//  Te_order (<=10)   : maximal order of temperature corrections. Te_order==0 means that the normal 
//                      thSZ formula is used.
//  betac_order (<=2) : maximal order of the kinematic corrections in the cluster frame. 
//                      betac_order == 0 means only the thSZ effect is considered.
//
//==================================================================================================
double compute_SZ_signal_asymptotic(double xo, 
                                    double Dtau, double Te, 
                                    double betac, double muc, 
                                    double betao, double muo, 
                                    int Te_order, int betac_order)
{
    double The=Te/const_me; // conversion keV --> The = k Te / me c^2
    double gammao=1.0/sqrt(1.0-betao*betao);
    double gammac=1.0/sqrt(1.0-betac*betac);
    
    // transformation of muc to mucc
    double mucc=(muc-betac)/(1.0-betac*muc);
    
    // transformation of xo to xc
    double xc=gammac*(1.0-betac*mucc)*gammao*xo*(1.0+betao*muo);
    
    // change to Nozawa 2006 convention for scattering optical depth
    if(!CNSN2012_convention) Dtau*=gammac*(1.0-betac*mucc);

    // sanity check
    if(Te_order>10){ Te_order=10; }
    if(betac_order>2){ betac_order=2; }
    
    return Dtau*compute_SZ_distortion_asymptotic(xc, The, betac, mucc, Te_order, betac_order,"all");
}


//==================================================================================================
//
//  Improved expansion of the Compton collision integral following the approach of 
//  Chluba, Nagai, Sazonov, Nelson, 2012 
//
//  The additional parameters are:
//  
//  Te_order (<=20)   : maximal order of temperature corrections. Te_order<=4 is not recommended.
//  betac_order (<=2) : maximal order of the kinematic corrections in the cluster frame. 
//                      betac_order == 0 means only the thSZ effect is considered.
//
//==================================================================================================
double compute_SZ_signal_CNSN_basis(double xo, 
                                    double Dtau, double Te, 
                                    double betac, double muc, 
                                    double betao, double muo, 
                                    int Te_order, int betac_order)
{
    double The=Te/const_me; // conversion keV --> The = k Te / me c^2
    double gammao=1.0/sqrt(1.0-betao*betao);
    double gammac=1.0/sqrt(1.0-betac*betac);
    
    // transformation of muc to mucc
    double mucc=(muc-betac)/(1.0-betac*muc);
    
    // transformation of xo to xc
    double xc=gammac*(1.0-betac*mucc)*gammao*xo*(1.0+betao*muo);
    
    // change to Nozawa 2006 convention for scattering optical depth
    if(!CNSN2012_convention) Dtau*=gammac*(1.0-betac*mucc);

    // sanity check
    if(Te_order>20){ Te_order=20; }
    if(betac_order>2){ betac_order=2; }
    
    return Dtau*compute_SZ_distortion_CNSN_basis(xc, The, betac, mucc, Te_order, betac_order,"all");
}


//==================================================================================================
//==================================================================================================