//==================================================================================================
//
// run_SZpack routine
// 
//==================================================================================================
//
// purpose: give some explicit examples of how to use SZpack. The code can be run with default 
// parameters that are set in the function
//
// int main(int narg, char *args[])
//
// below. Alternatively one can call it with a start file.
//
//==================================================================================================
//
// Author: Jens Chluba (CITA, University of Toronto)
//
// first implementation: May 2012
// last modification: June 2012
//
//==================================================================================================


//==================================================================================================
// Standards
//==================================================================================================
#include <stdio.h>
#include <iostream>
#include <fstream>
#include <string>
#include <ctime>
#include <iomanip>
#include <cmath>
#include <limits.h>
#include <vector>

//==================================================================================================
// required libs
//==================================================================================================
#include "SZpack.h"
#include "physical_consts.h"
#include "routines.h"

//==================================================================================================
//
// namespaces
//
//==================================================================================================
using namespace std;

bool show_mess=1;

//==================================================================================================
//
// print message to screen
//
//==================================================================================================
void print_message(string mess)
{
    if(show_mess)
    {
        cout << "\n ====================================================="
        << "===================================================== " << endl;
        cout << " || " << mess << endl;
        cout << " ====================================================="
        << "===================================================== " << endl << endl;   
    }
    
    return;
}

//==================================================================================================
//
// header for file
//
//==================================================================================================
void output_header(ofstream &ofile, 
                   double xmin, double xmax, int np, 
                   double Dtau, double Te, 
                   double betac, double muc, 
                   double betao, double muo, 
                   double eps_Int, int Te_order, int beta_order,
                   string mode)
{
    ofile.precision(6);

    ofile << "########################################################################\n#" << endl;
    ofile << "# SZ signal computed with " << SZpack_version << endl;
    ofile << "# Runmode: " << mode << endl;
    ofile << "#\n# Parameters:" << endl;
    ofile << "# xmin= " << xmin << " xmax= " << xmax << " np= " << np << endl;
    ofile << "# Dtau= " << Dtau << endl;
    ofile << "# Te= " << Te << "keV (The ~ " << Te/const_me << ")" << endl;
    ofile << "# betac= " << betac << " muc= " << muc << endl;
    ofile << "# betao= " << betao << " muo= " << muo << endl;
    ofile << "#\n# Runmode specific parameters:\n# eps= " << eps_Int << endl;
    ofile << "# Te_order= " << Te_order << " beta_order= " << beta_order << endl;
    ofile << "#\n######################################################################## " << endl;
    ofile << "#\n# Output format: x = (h nu/k T0) | x^3 Dn(x) " << endl;
    ofile << "#\n######################################################################## " << endl;
    
    
    return;
}

//==================================================================================================
//
// compute distortion
//
//==================================================================================================
void output_SZ_distortion_5D(string fname, 
                             double xmin, double xmax, int np, 
                             double Dtau, double Te, 
                             double betac, double muc, 
                             double betao, double muo, 
                             double eps_Int=1.0e-4)
{
    print_message("Carrying out 5D integral");
    
    vector<double> xa(np);
    init_xarr(xmin, xmax, &xa[0], np, 1, 0); // change 1 --> 0 to have linear grid in x
    
    ofstream ofile(fname.c_str());
    
    output_header(ofile, xmin, xmax, np, Dtau, Te, betac, muc, betao, muo, 
                  eps_Int, 0, 0,
                  "5D integral");
    
    ofile.precision(16);

    for(int k=0; k<np; k++)
    {
        double dum=compute_SZ_signal_5D(xa[k], Dtau, Te, 
                                        betac, muc, betao, muo, 
                                        eps_Int);
        
        ofile << xa[k] << " " << dum*pow(xa[k], 3) << endl;
        cout  << " x= " << xa[k] << " " << dum*pow(xa[k], 3) << endl;
    }
        
    ofile.close();
        
    return;
}

//==================================================================================================
void output_SZ_distortion_3D(string fname, 
                             double xmin, double xmax, int np, 
                             double Dtau, double Te, 
                             double betac, double muc, 
                             double betao, double muo, 
                             double eps_Int=1.0e-4)
{
    print_message("Carrying out 3D integral");

    vector<double> xa(np);
    init_xarr(xmin, xmax, &xa[0], np, 1, 0); // change 1 --> 0 to have linear grid in x
    
    ofstream ofile(fname.c_str());

    output_header(ofile, xmin, xmax, np, Dtau, Te, betac, muc, betao, muo, 
                  eps_Int, 0, 0,
                  "3D integral");

    ofile.precision(16);

    for(int k=0; k<np; k++)
    {
        double dum=compute_SZ_signal_3D(xa[k], Dtau, Te, 
                                        betac, muc, betao, muo, 
                                        eps_Int);
        
        ofile << xa[k] << " " << dum*pow(xa[k], 3) << endl;
        cout  << " x= " << xa[k] << " " << dum*pow(xa[k], 3) << endl;
    }
    
    ofile.close();
    
    return;
}

//==================================================================================================
void output_SZ_distortion_asym(string fname, 
                               double xmin, double xmax, int np, 
                               double Dtau, double Te, 
                               double betac, double muc, 
                               double betao, double muo,
                               int Te_order, int beta_order)
{
    print_message("Using asymptotic expansion of collision integral");
    
    vector<double> xa(np);
    init_xarr(xmin, xmax, &xa[0], np, 1, 0); // change 1 --> 0 to have linear grid in x
    
    ofstream ofile(fname.c_str());
    
    output_header(ofile, xmin, xmax, np, Dtau, Te, betac, muc, betao, muo, 
                  0, Te_order, beta_order,
                  "asymptotic");
    
    ofile.precision(16);
    
    for(int k=0; k<np; k++)
    {
        double dum=compute_SZ_signal_asymptotic(xa[k], Dtau, Te, 
                                                betac, muc, betao, muo, 
                                                Te_order, beta_order);
        
        ofile << xa[k] << " " << dum*pow(xa[k], 3) << endl;
        cout  << " x= " << xa[k] << " " << dum*pow(xa[k], 3) << endl;
    }
    
    ofile.close();
    
    return;
}

//==================================================================================================
void output_SZ_distortion_CNSN(string fname, 
                               double xmin, double xmax, int np, 
                               double Dtau, double Te, 
                               double betac, double muc, 
                               double betao, double muo,
                               int Te_order, int beta_order)
{
    print_message("Using improved basis of CNSN 2012");
    
    vector<double> xa(np);
    init_xarr(xmin, xmax, &xa[0], np, 1, 0); // change 1 --> 0 to have linear grid in x
    
    ofstream ofile(fname.c_str());
    
    output_header(ofile, xmin, xmax, np, Dtau, Te, betac, muc, betao, muo, 
                  0, Te_order, beta_order,
                  "CNSN-basis");
    
    ofile.precision(16);
    
    for(int k=0; k<np; k++)
    {
        double dum=compute_SZ_signal_CNSN_basis(xa[k], Dtau, Te, 
                                                betac, muc, betao, muo, 
                                                Te_order, beta_order);
        
        ofile << xa[k] << " " << dum*pow(xa[k], 3) << endl;
        cout  << " x= " << xa[k] << " " << dum*pow(xa[k], 3) << endl;
    }
    
    ofile.close();
    
    return;
}

//==================================================================================================
//
// read parameters
//
//==================================================================================================
void read_parameters(string fname, 
                     double &xmin, double &xmax, int &np, 
                     double &Dtau, double &Te, 
                     double &betac, double &muc, 
                     double &betao, double &muo,
                     int &Te_order, int &beta_order, double &eps_Int,
                     string &outpath, string &add)
{
    ifstream ifile(fname.c_str());
    
    ifile >> xmin; ifile >> xmax; ifile >> np;
    
    ifile >> Dtau; ifile >> Te;

    ifile >> betac; ifile >> muc;
    ifile >> betao; ifile >> muo;

    ifile >> Te_order; ifile >> beta_order; ifile >> eps_Int;
    
    ifile >> outpath; ifile >> add;
    
    // sanity checks
    if(xmin<0.1 || xmax>50.0){ cerr << " change x-range " << endl; exit(0); }
    if(np<=0){ cerr << " change number of frequency points " << endl; exit(0); }

    if(Dtau<0){ cerr << " change Dtau " << endl; exit(0); }
    if(Te>50.0 || Te<0){ cerr << " change Te " << endl; exit(0); }   

    if(betac<0 || betac>0.1){ cerr << " check betac " << endl; exit(0); }
    if(muc<-1.0 || muc>1){ cerr << " check muc " << endl; exit(0); }
    if(betao<0 || betao>0.1){ cerr << " check betao " << endl; exit(0); }
    if(muo<-1.0 || muo>1){ cerr << " check muo " << endl; exit(0); }

    if(eps_Int<1e-8 || eps_Int>0.001){ cerr << " check eps_Int " << endl; exit(0); }

    return; 
}
 
//==================================================================================================
//
// main call of routine. Here parameters are handled.
//
//==================================================================================================
int main(int narg, char *args[])
{
    //==================================================================
    // example parameters
    //==================================================================
    double xmin=0.1;
    double xmax=30.0;
    int np=50;
    //
    double Dtau=0.01;
    double Te=10.0;
    //
    double betac=0.01;
    double muc=1.0;
    //
    double betao=0.001241;
    double muo=0.0;
    //
    double eps_Int=1.0e-4;
    int Te_order=10;
    int beta_order=2;
    //
    string outpath=SZPACKDIR+"./outputs/";
    string add=".dat";
    string mode="3D";
    
    //==================================================================
    // startup with different number of arguments
    //==================================================================
    if(narg==1){}
    
    else if(narg==2) mode=args[narg-1];
    
    else if(narg==3)
    {
        mode=args[narg-2];
        string fname=args[narg-1];
        
        print_message("Reading parameters from parameter file: " + fname);
        
        read_parameters(fname, xmin, xmax, np, Dtau, Te, 
                        betac, muc, betao, muo,
                        Te_order, beta_order, eps_Int, outpath, add);
    }

    else{ cerr << " Too many/few parameters " << endl; exit(1); }
        
    //==================================================================
    // uncomment this to use the convention of Nozawa et al. 2006 for 
    // the optical depth. Interpretation of the SZ signal is different
    // in this case. The convention of Chluba et al., 2012 is used  
    // otherwise, which explicitly imposes the Lorentz-invariance of  
    // the scattering optical depth.
    //==================================================================
    //use_Nozawa2006_convention();
    //add=".Nozawa_conv"+add;

    //==================================================================
    // call in the different runmodes
    //==================================================================
    if(mode=="5D")
    {
        output_SZ_distortion_5D(outpath+"SZ_Integral.5D"+add, 
                                xmin, xmax, np, Dtau, Te, 
                                betac, muc, betao, muo, 
                                eps_Int);
    }
    
    else if(mode=="3D")
    {
        output_SZ_distortion_3D(outpath+"SZ_Integral.3D"+add, 
                                xmin, xmax, np, Dtau, Te, 
                                betac, muc, betao, muo, 
                                eps_Int);
    }
    
    else if(mode=="ASYM")
    {
        output_SZ_distortion_asym(outpath+"SZ_asymptotic"+add, 
                                  xmin, xmax, np, Dtau, Te, 
                                  betac, muc, betao, muo, 
                                  Te_order, beta_order);
    }
    
    else if(mode=="CNSN")
    {
        output_SZ_distortion_CNSN(outpath+"SZ_CNSN_basis"+add, 
                                  xmin, xmax, np, Dtau, Te, 
                                  betac, muc, betao, muo, 
                                  Te_order, beta_order);
    }
    
    else cerr << " Unknown runmode: " << mode << endl;
    
    return 0;
}

//==================================================================================================
//==================================================================================================
