//==================================================================================================
//
// This program allows computing the thermal SZ and kinematic effect using the improved basis of 
// Chluba, Nagai, Sazonov & Nelson, 2012. Up to 20th order temperature corrections are allowed. 
// Kinematic corrections can be included up to second order in the clusters peculiar velocity. 
// The SZ signals are computed in the cluster frame. For x <30, 5keV < Te < 25keV and beta <0.01  
// the obtained SZ signal should be accurate at the level of 0.001% when 20 temperature terms are 
// included. For Te < 5keV the asymptotic expansions should be used.
//
//==================================================================================================
//
// Author: Jens Chluba  (CITA, University of Toronto)
//
// first implementation: April 2012
// last modification   : June 2012
//
//==================================================================================================

//==================================================================================================
// Standards
//==================================================================================================
#include <stdio.h>
#include <iostream>
#include <fstream>
#include <string>
#include <ctime>
#include <iomanip>
#include <cmath>
#include <limits.h>
#include <vector>

//==================================================================================================
// required libs
//==================================================================================================
#include "physical_consts.h"
#include "routines.h"
#include "Relativistic_MB.h"
#include "nPl_derivatives.h"
#include "SZ_CNSN_basis.h"

//==================================================================================================
//
// namespaces
//
//==================================================================================================
using namespace std;


bool basis_loaded=0;
const string SZ_functions_path=SZPACKDIR+"./src/database/";

const double Tref_basis=0.03; // this is the reference temperature that was used to compute basis;
const int mmax=20;            // maximal temperature order

//==================================================================================================
//
// SZ effect using interpolation of derivative terms
//
//==================================================================================================
vector<int> spline_mem_indices_th;
vector<int> spline_mem_indices_d;
vector<int> spline_mem_indices_q;
vector<int> spline_mem_indices_m;

//==================================================================================================
void load_expansions(string fname, vector<int> &spline_mem_indices, int Torder)
{
    load_data_from_file(fname, Torder+2, spline_mem_indices, 1, 0);
    
    return;
}

//==================================================================================================
// generic function for interpolation (x^3 Dn was used)
//==================================================================================================
double DI_SZ_splines(double x, double The, vector<int> &spline_mem_indices, int Torder)
{
    double lx=log(x);
    double r=calc_spline_JC(lx, spline_mem_indices[0]);

    for(int i=2; i<=Torder; i++) 
        r+=pow(The/Tref_basis-1.0, i-1)*calc_spline_JC(lx, spline_mem_indices[i-1])/factorial(i-1);
    
    r*=norm_df_RM_dTheta(The)/norm_df_RM_dTheta(Tref_basis);

    return r;     
}

//==================================================================================================
// add different terms
//
// mode == "monopole"      --> only scattering of monopole without second order kinematic corr
// mode == "dipole"        --> only scattering of dipole     (first order kinematic correction)
// mode == "quadrupole"    --> only scattering of quadrupole (second order kinematic correction)
// mode == "monopole_corr" --> only scattering of second order kinematic correction to monopole
// mode == "all"           --> all terms added
// mode == "kin"           --> only kinematic terms
//
//==================================================================================================
double compute_SZ_distortion_CNSN_basis(double x, 
                                        double The, double betac, double muc, 
                                        int Te_order, int betac_order,
                                        string mode)
{
    double r=0.0, rb;
    
    if(!basis_loaded)
    {
        cout << " Loading CNSN 2012 basis functions. This will only be done once. " << endl;
        cout << endl;
        
        load_expansions(SZ_functions_path+"SZ_basis.3D.The_0.03.monopole.dat", 
                        spline_mem_indices_th, mmax);
        load_expansions(SZ_functions_path+"SZ_basis.3D.The_0.03.dipole.dat", 
                        spline_mem_indices_d, mmax);
        load_expansions(SZ_functions_path+"SZ_basis.3D.The_0.03.quadrupole.dat", 
                        spline_mem_indices_q, mmax);
        load_expansions(SZ_functions_path+"SZ_basis.3D.The_0.03.monopole_corr.dat", 
                        spline_mem_indices_m, mmax);    
        basis_loaded=1;
        
        cout << endl;
    }
        
    Te_order++;

    if(betac_order>0)
    {
        if(mode=="dipole" || mode=="all" || mode=="kin") 
        {
            rb=DI_SZ_splines(x, The, spline_mem_indices_d, Te_order);
            rb+=pow(x, 3)*x*exp(-x)/pow(one_minus_exp_mx(x), 2);
            r+=betac*muc*rb;
        }

        if(betac_order>1)
        {
            if(mode=="monopole_corr" || mode=="all" || mode=="kin") 
            {
                rb=DI_SZ_splines(x, The, spline_mem_indices_m, Te_order);
                r+=betac*betac*rb;
            }  
            
            if(mode=="quadrupole" || mode=="all" || mode=="kin") 
            {
                rb=DI_SZ_splines(x, The, spline_mem_indices_q, Te_order);
                rb-=pow(x, 3)*x*x*exp(-x)*(exp(-x)+1.0)/pow(one_minus_exp_mx(x), 3)/3.0;
                r+=betac*betac*(3.0*muc*muc-1.0)*0.5*rb;
            }  
        }
    }
    
    if(mode=="monopole" || mode=="all") 
    {
        r+=DI_SZ_splines(x, The, spline_mem_indices_th, Te_order);
    }
        
    return r/pow(x, 3);     
}

//==================================================================================================
//==================================================================================================
